// 67d7842dbbe25473c3c32b93c0da8047785f30d78e8a024de1b57352245f9689
/*******************************************************************************
 * Copyright (c) 2010, 2015 Wind River Systems, Inc. and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * Contributors:
 *     Wind River Systems - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * Modified by Xilinx,Inc
 * Copyright 2022 Xilinx, Inc. All Rights Reserved. Modifications
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 *******************************************************************************/

package org.eclipse.tcf.internal.dap;

import org.eclipse.lsp4j.debug.Breakpoint;
import org.eclipse.lsp4j.debug.BreakpointEventArguments;
import org.eclipse.lsp4j.debug.BreakpointEventArgumentsReason;
import org.eclipse.tcf.internal.dap.actions.TCFAction;
import org.eclipse.tcf.internal.dap.breakpoint.BreakpointInfo;
import org.eclipse.tcf.internal.dap.breakpoint.TCFBreakpointHandler;
import org.eclipse.tcf.protocol.Protocol;
import org.eclipse.tcf.services.IBreakpoints;

import java.util.*;

/**
 * This class monitors breakpoints status on TCF debug targets and calls ICBreakpoint.incrementInstallCount() or
 * ICBreakpoint.decrementInstallCount() when breakpoint status changes.
 */
@SuppressWarnings("restriction")
class TCFBreakpointStatusListener {

    /** Ref count attribute for foreign breakpoints */
    private static final String ATTR_REFCOUNT = "org.eclipse.tcf.cdt.refcount";

    private class BreakpointListener implements ITCFBreakpointListener {

        private final TCFLaunch launch;
        private final TCFBreakpointsStatus status;
//        private final Map<String,ICBreakpoint> installed = new HashMap<String,ICBreakpoint>();
        private final Map<String, BreakpointInfo> installed = new HashMap<>();
        private final Map<String, Integer> foreignIds = new HashMap<>();
        private final Set<String> deleted = new HashSet<>();

        BreakpointListener(TCFLaunch launch) {
            this.launch = launch;
            status = launch.getBreakpointsStatus();
            status.addListener(this);
            bp_listeners.put(launch, this);
//            DebugPlugin.getDefault().getBreakpointManager().addBreakpointListener(this);
            for (String id : status.getStatusIDs()) breakpointStatusChanged(id);
        }

        public void breakpointStatusChanged(String id) {
            BreakpointInfo bp = bp_model.getBreakpoint(id);
            updateStatus(id, bp);
            if (bp == null) {
                createOrUpdateBreakpoint(id);
            }
        }

/*
        public void breakpointAdded(IBreakpoint breakpoint) {}
        public void breakpointChanged(IBreakpoint breakpoint, IMarkerDelta delta) {
            updateBreakpoint(breakpoint, false);
        }
        public void breakpointRemoved(final IBreakpoint breakpoint, IMarkerDelta delta) {
            updateBreakpoint(breakpoint, true);
        }

        private void updateBreakpoint(IBreakpoint breakpoint, final boolean removed) {
            try {
                IMarker marker = breakpoint.getMarker();
                if (marker == null || !marker.exists()) return;
                if (TCFBreakpointsModel.isLocal(marker)) return;
                final String marker_id = TCFBreakpointsModel.getBreakpointID(breakpoint);
                Protocol.invokeLater(new Runnable() {
                    public void run() {
                        if (removed) {
                            foreign.remove(marker_id);
                        }
                        createOrUpdateBreakpoint(marker_id);
                    }
                });
            }
            catch (CoreException e) {
            }
        }
*/

        private void updateStatus(String id, BreakpointInfo bp) {
            if (bp != null) {
                boolean ok = false;
//                ICBreakpoint cbp = (ICBreakpoint)bp;
                Map<String,Object> map = status.getStatus(id);
                if (map != null) {
                    @SuppressWarnings("unchecked")
                    Collection<Map<String,Object>> list = (Collection<Map<String,Object>>)map.get(IBreakpoints.STATUS_INSTANCES);
                    if (list != null) {
                        for (Map<String,Object> m : list) {
                            if (m.get(IBreakpoints.INSTANCE_ERROR) == null) ok = true;
                        }
                    }
                }
                if (ok && installed.get(id) == null) {
                    installed.put(id, bp);
//                    incrementInstallCount(cbp);
                }
                else if (!ok && installed.get(id) == bp) {
                    installed.remove(id);
//                    decrementInstallCount(cbp);
                }
                else {
//                    updateTCFStamp(cbp);
                }
            }
//            else if (bp instanceof TCFBreakpoint) {
//                updateStatus((TCFBreakpoint)bp);
//            }
        }

        public void breakpointRemoved(String id) {
            BreakpointInfo cbp = installed.remove(id);
            if (cbp != null) {
//                decrementInstallCount(cbp);
            }
            if (foreignIds.containsKey(id)) {
                foreignIds.remove(id);
                deleteTransientBreakpoint(id);
            }
        }

        public void breakpointChanged(String id) {
            createOrUpdateBreakpoint(id);
        }

        public void dispose() {
//            DebugPlugin.getDefault().getBreakpointManager().removeBreakpointListener(this);
//            for (ICBreakpoint cbp : installed.values()) {
//                decrementInstallCount(cbp);
//            }
            installed.clear();
            for (String id : foreignIds.keySet()) {
                deleteTransientBreakpoint(id);
            }
            foreignIds.clear();
        }

/*
        private void incrementInstallCount(final ICBreakpoint cbp) {
            Job job = new WorkspaceJob("Increment Breakpoint Install Count") {
                { setSystem(true); }
                @Override
                public IStatus runInWorkspace(IProgressMonitor monitor) throws CoreException {
                    try {
                        cbp.incrementInstallCount();
                        doUpdateTCFStamp(cbp);
                    }
                    catch (CoreException e) {
                        // ignore expected race condition with marker deletion
                    }
                    return Status.OK_STATUS;
                }
            };
            job.setRule(cbp.getMarker().getResource());
            job.setPriority(Job.SHORT);
            job.setSystem(true);
            job.schedule();
        }

        private void decrementInstallCount(final ICBreakpoint cbp) {
            Job job = new WorkspaceJob("Decrement Breakpoint Install Count") {
                { setSystem(true); }
                @Override
                public IStatus runInWorkspace(IProgressMonitor monitor) throws CoreException {
                    try {
                        cbp.decrementInstallCount();
                        doUpdateTCFStamp(cbp);
                    }
                    catch (CoreException e) {
                        // ignore expected race condition with marker deletion
                    }
                    return Status.OK_STATUS;
                }
            };
            job.setRule(cbp.getMarker().getResource());
            job.setPriority(Job.SHORT);
            job.setSystem(true);
            job.schedule();
        }
*/

/*
        private void updateTCFStamp(final ICBreakpoint cbp) {
            Job job = new WorkspaceJob("Update C Breakpoint") {
                { setSystem(true); }
                @Override
                public IStatus runInWorkspace(IProgressMonitor monitor) throws CoreException {
                    try {
                        doUpdateTCFStamp(cbp);
                    }
                    catch (CoreException e) {
                        // ignore expected race condition with marker deletion
                    }
                    return Status.OK_STATUS;
                }
            };
            job.setRule(cbp.getMarker().getResource());
            job.setPriority(Job.SHORT);
            job.setSystem(true);
            job.schedule();
        }

        private void doUpdateTCFStamp(ICBreakpoint cbp) throws CoreException {
            IMarker marker = cbp.getMarker();
            if (marker != null && marker.exists()) {
                marker.setAttribute(TCFBreakpointsModel.ATTR_TCF_STAMP, "true");
            }
        }
*/

/*
        private void updateStatus(final TCFBreakpoint tbp) {
            Job job = new WorkspaceJob("Update Breakpoint Status") {
                @Override
                public IStatus runInWorkspace(IProgressMonitor monitor) throws CoreException {
                    try {
                        tbp.notifyStatusChaged();
                    }
                    catch (CoreException e) {
                        // ignore expected race condition with marker deletion
                    }
                    return Status.OK_STATUS;
                }
            };
            job.setRule(tbp.getMarker().getResource());
            job.setPriority(Job.SHORT);
            job.setSystem(true);
            job.schedule();
        }
*/

        private void createOrUpdateBreakpoint(final String id) {
            if (id.startsWith(TCFAction.STEP_BREAKPOINT_PREFIX)) return;
            Map<String,Object> properties = status.getProperties(id);
            if (properties == null) return;
            if (TCFBreakpointsModel.isLocal(properties)) return;
            final boolean create = !foreignIds.containsKey(id);
            if (create) {
                foreignIds.put(id, breakpointHandler.nextBreakpointId());
            }
            if (deleted.remove(id)) return;

            BreakpointInfo bpInfo = bp_model.createForeignBreakpointInfo(id, properties, foreignIds.get(id));
            Breakpoint[] breakpoints = breakpointHandler.createBreakpointsFromInfo(new BreakpointInfo[] {bpInfo});
            if (breakpoints != null && breakpoints.length > 0) {
                BreakpointEventArguments eventArgs = new BreakpointEventArguments();
                eventArgs.setBreakpoint(breakpoints[0]);
                eventArgs.setReason(create ? BreakpointEventArgumentsReason.NEW : BreakpointEventArgumentsReason.CHANGED);
                launch.getClient().breakpoint(eventArgs);
            }

            if (create) {
                Protocol.invokeAndWait(new Runnable() {
                    public void run() {
                        bp_model.setBreakpoint(id, bpInfo);
                        updateStatus(id, bpInfo);
                    }
                });
            }
        }

        private void deleteTransientBreakpoint(final String id) {
            BreakpointEventArguments eventArgs = new BreakpointEventArguments();
            Breakpoint breakpoint = new Breakpoint();
            breakpoint.setId(foreignIds.get(id));
            eventArgs.setBreakpoint(breakpoint);
            eventArgs.setReason(BreakpointEventArgumentsReason.REMOVED);
            launch.getClient().breakpoint(eventArgs);
            // Since breakpoint object is created by a another background job after reading data from remote peer,
            // this job can be running before the job that creates the object.
            // We need to remember ID of the breakpoint that became obsolete before it was fully created.
            deleted.add(id);
        }

/*
        private ISchedulingRule getBreakpointAccessRule() {
            IResource resource = ResourcesPlugin.getWorkspace().getRoot();
            ISchedulingRule rule = ResourcesPlugin.getWorkspace().getRuleFactory().markerRule(resource);
            if (rule == null) {
                // In Eclipse 3.6.2, markerRule() always returns null,
                // causing race condition, a lot of crashes and corrupted data.
                // Using modifyRule() instead.
                rule = ResourcesPlugin.getWorkspace().getRuleFactory().modifyRule(resource);
            }
            return rule;
        }
*/
    }

    private TCFBreakpointHandler breakpointHandler;
    public void setBreakpointHandler(TCFBreakpointHandler breakpointHandler) {
        this.breakpointHandler = breakpointHandler;
    }

    private final TCFModelManager.ModelManagerListener launch_listener = new TCFModelManager.ModelManagerListener() {

        public void onConnected(TCFLaunch launch, TCFModel model) {
            assert bp_listeners.get(launch) == null;
            if (launch.getBreakpointsStatus() != null) new BreakpointListener(launch);
        }

        public void onDisconnected(TCFLaunch launch, TCFModel model) {
            BreakpointListener l = bp_listeners.remove(launch);
            if (l != null) l.dispose();
        }
    };

    private final TCFModelManager model_manager;
    private final TCFBreakpointsModel bp_model;
    private final Map<TCFLaunch,BreakpointListener> bp_listeners;

    TCFBreakpointStatusListener() {
        bp_model = TCFBreakpointsModel.getBreakpointsModel();
        model_manager = TCFModelManager.getInstance();
        model_manager.addListener(launch_listener);
        bp_listeners = new HashMap<TCFLaunch,BreakpointListener>();
        // handle already connected launches
        for (TCFModel model : model_manager.getModels()) {
            TCFLaunch launch = model.getLaunch();
            if (launch.isConnected()) launch_listener.onConnected(launch, model);
        }
    }

    void dispose() {
        model_manager.removeListener(launch_listener);
        for (BreakpointListener l : bp_listeners.values()) l.dispose();
        bp_listeners.clear();
    }
}
